/* $Workfile: DspDemo2.c $   $Revision: 24 $    $Modtime: 9/27/04 11:46a $  */
/*--------------------------------------------------------------------------*\

  Odin TeleSystems Inc.

  SAA-1007-1 OTX DSP C54x Software Development Kit (SDK)

  DspDemo2

  This demo illustrates how to send and receive serial user data. It receives
  u-law encoded voice samples. It loops these samples back in the transmit
  direction after performing a gain adjustment. The gain adjustment value
  is selected by passing an IO Control code to this application.

  Use together with Host application DspDemo2.c

\*--------------------------------------------------------------------------*/

#include <string.h>   // for memset()


#ifdef CHIP_54CST
// Defines CST channel structure.
// Defines a global variable Ch0 to be used for single channel applications.
// This channel is also a reference address for S-register description.
#include "Framework\CSTChannel.h"

/*
  Use all available patches (except the AT Parser patch because
  usually flex applications do not use the AT Parser,
  except the default settings patch because it needs more data RAM
  and except the uart driver patch because it doesn't affect functionality)
*/
#define APPLY_ATPARSER_PATCH    0
#define APPLY_UARTDRV_PATCH     0
#define APPLY_MODEM_PATCH       0
#define APPLY_VAD_PATCH         0
#define APPLY_SETTINGS_PATCH    0
#define APPLY_CSTSERVICE_PATCH  0
#define APPLY_ATOMIC_PATCH      0
#include "Patch\CSTPatch.h"
#endif // CHIP_54CST


#include "Dsp54Api.h"

#include <Dsp54Ioctl.h>

#include "Demo2Io.h"

#if 0
#ifdef MCBSP_SUPPORT
#include <csl.h>
#include <csl_mcbsp.h>
#include <csl_dma.h>
#include <csl_irq.h>
#endif
#endif

/*--------------------------------------------------------------------------*\
  Macros
\*--------------------------------------------------------------------------*/
#define DEMO_MAX_WORDS_PER_FRAME 64
#define DEMO_DATA_BUF_SIZE  25

#if 1 // Temporary
extern unsigned short g_debugData[8];
extern unsigned short g_debugBuf[32];
#endif

unsigned short g_nFrameCounter = 0;

/*--------------------------------------------------------------------------*\
  Private data
\*--------------------------------------------------------------------------*/

unsigned short g_hHandle = OTXDSP_INVALID_HANDLE;  // Handle passed to us from the host user application

static unsigned short g_sqFrameSample[DEMO_MAX_WORDS_PER_FRAME];  // Temporary storage area for one frame

static short g_law = OTXDSP_ULAW;

static short g_nTxGain = 0;

static short g_bRxEvenTimeSlot = OTXDSP_TRUE;   // Even or odd timeslot for the receiver
static short g_nRxWordSlot = 0;       // The "Word slot" (0-15) within a frame for the receiver

static short g_bTxEvenTimeSlot = OTXDSP_TRUE;   // Even or odd timeslot for the transmitted data
static short g_nTxWordSlot = 0;       // The "Word slot" (0-15) within a frame for the transmitted data

static short g_bRxEnabled = OTXDSP_FALSE;
static short g_bTxEnabled = OTXDSP_FALSE;

static short g_sqRxData[DEMO_DATA_BUF_SIZE];
static short g_sqTxData[DEMO_DATA_BUF_SIZE];
static short g_iTxData = 0;
static short g_iRxData = 0;

static short g_bStreamRxData = OTXDSP_FALSE;
static short g_bStreamFullFrames = OTXDSP_FALSE;
static short g_bConstValMode = OTXDSP_FALSE;

Demo2NotifyGetDebugDataS g_debugDataS;

static volatile long  g_blinkCount      = 0;
static volatile long  g_blinkCountMax   = 100000L;
static volatile short g_blinkXfVal      = OTXDSP_FALSE;

#pragma DATA_SECTION(g_sqTestData,"extdmem");
unsigned short g_sqTestData[256];


#ifdef CHIP_54CST
  //Prevent improper compilation
  asm("__sys_memory .usect  \".sysmem\",0");
  asm("__STACK_BEGINNING .usect \".stack\",0");
#endif

/*--------------------------------------------------------------------------*\
  Function Prototypes
\*--------------------------------------------------------------------------*/
static void ProcessRxSamples(void);
static void GenerateTxSamples(void);
static void UnpackFrame(unsigned short *pFrame);
static void PackFrame(unsigned short *pFrame);
static void Blink(void);

#ifdef CHIP_54CST
/*--------------------------------------------------------------------------*\
  @func
  Implementation of the main callback function called
  from CST. Each CST Action-oriented program should
  implement this method. Almost all data and info go through this
  callback routine.

  pChannel - Pointer to the channel (the particular CST instance)
  CSTExternalMsgEvent - Event info (message code)
  Data - Depending on the event, it can be data byte or length of data bytes
    in pData buffer or nothing
  pData - Depending on the event, it can be data buffer or nothing
\*--------------------------------------------------------------------------*/
static bool OtxDspCstCallback(
  tCSTChannel *pChannel,
  tCSTExternalMsgEvent CSTExternalMsgEvent,
  int Data,
  int16 *pData
)
{
  switch (CSTExternalMsgEvent) {
    //////////////////
    //USER'S CODE...//
    //////////////////
  }
  return 1;
}
#endif // CHIP_54CST


/*--------------------------------------------------------------------------*\
  @func
    main() is called from _c_int00 which is the entry point for the boot loader.
    _c_int00 will first initialize all global data before running main().
\*--------------------------------------------------------------------------*/
main() {

#ifdef CHIP_54CST
  // Processor boot init.
  CST_DSPInit();

  // CST internal data sections init.
  CST_bssInit();

  // CST Framework init.
  CSTAction_Init(&Ch0, 0, OtxDspCstCallback);
#endif // CHIP_54CST

  // Run Standard OTX DSP SDK API Initializations
  OtxDspInit();

#ifdef CHIP_54CST
  // Apply all chosen CST patches
  //CSTPatchApply();

  OtxDspLawInit();
#endif // CHIP_54CST

  // Initialize application specific data
  memset(&g_debugDataS, 0, sizeof(Demo2NotifyGetDebugDataS));

  // Loop forever
  for(;;) {
#if 0
#ifdef CHIP_54CST
    CSTPatchServiceProcess(&Ch0);
#endif // CHIP_54CST
#endif    
    // check if we have RX samples available, process them. E.g. run detection algorithms on them
    if (g_bGotRxSerialPortInterrupt) {  // g_bGotRxSerialPortInterrupt updated by the interrupt service routine
      g_bGotRxSerialPortInterrupt = OTXDSP_FALSE;
      OtxDspServiceRxSerialPort();
    }

    // check if more samples are needed to be transmitted, generate more samples.
    if (g_bGotTxSerialPortInterrupt) {  // g_bGotTxSerialPortInterrupt updated by the interrupt service routine
      g_bGotTxSerialPortInterrupt = OTXDSP_FALSE;
      OtxDspServiceTxSerialPort();
    }

    // check if the host has requested our attention. If so, service his request
    if (g_bGotHostPortInterrupt) {  // g_bGotHostPortInterrupt updated by the interrupt service routine
      g_bGotHostPortInterrupt = OTXDSP_FALSE;
      OtxDspServiceIoControl();
    }
    
    // Blink the heartbeat LED to show that the DSP is up and running
    Blink();

  }  /* for */

}


/*--------------------------------------------------------------------------*\
  @func
    Blink the heart-beat LED
\*--------------------------------------------------------------------------*/
void Blink(
  void
)
{

  if (g_blinkCount++ >= g_blinkCountMax) {
    g_blinkCount = 0;

    if (g_blinkXfVal) {
      g_blinkXfVal = OTXDSP_FALSE;
      C54X_CLR_XF_BIT;   // Turn LED on
    } else {
      g_blinkXfVal = OTXDSP_TRUE;
      C54X_SET_XF_BIT;   // Turn LED off
    }
  }

}

#if 0 // Just for testing
#pragma CODE_SECTION(Foo,"extpmem");
short Foo(void)
{
  short i = 0;
  short j = 1;
  i = i + j;
  i++;
  i++;
  i++;
  return i;
}
#endif


/*--------------------------------------------------------------------------*\
    Call-back routine called by the Interrupt Service Routine when the
    host has issued a command to the HPI and requests our attention.

    This function is called from the idle loop in main()
    TO BE RE-IMPLEMENTED BY THE USER.
\*--------------------------------------------------------------------------*/
void OtxDspServiceIoControl(
  void
)
{
  //unsigned short nDataLen      = *((unsigned short *) OTXDSP_SDK_IOCTL_DATA_LEN_ADDR);
  unsigned short hHandle       = *((unsigned short *) OTXDSP_SDK_IOCTL_HANDLE_ADDR);
  unsigned short nTaskRef      = *((unsigned short *) OTXDSP_SDK_IOCTL_TASKREF_ADDR);
  //short          bDataComplete = *((unsigned short *) OTXDSP_SDK_IOCTL_CODE_ADDR) & OTXDSP_SDK_MASK_IO_DATA_COMPLETE;
  //short          bDataPadded   = *((unsigned short *) OTXDSP_SDK_IOCTL_CODE_ADDR) & OTXDSP_SDK_MASK_IO_DATA_PADDED;
  unsigned short nCode         = *((unsigned short *) OTXDSP_SDK_IOCTL_CODE_ADDR) & OTXDSP_SDK_MASK_IO_CODE;

  // Standard SDK IO Control codes
  if (nCode == OTXDSP_SDK_IO_CREATE) {
    if (g_hHandle == OTXDSP_INVALID_HANDLE) {
      // The handle is not set
      g_hHandle = hHandle;
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);
    } else {
      // This demo only supports one instance of the logical device. Return an error code.
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_OUT_OF_RESOURCES);
    }

  } else if (nCode == OTXDSP_SDK_IO_CLOSE) {
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_ENABLE) {
    g_bRxEnabled = OTXDSP_TRUE;
    g_bTxEnabled = OTXDSP_TRUE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_DISABLE) {
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_CONNECT) {
    OtxDspSdkIoInConnectS *pInS = (OtxDspSdkIoInConnectS *)OTXDSP_SDK_IOCTL_DATA_ADDR;
    short i;
    // Save these values in the debug struct. To be retrieved when the DSPDEMO1_IOCTL_GET_DEBUG_DATA control is issued
    g_debugDataS.nData0     = pInS->nDirection;

    if (pInS->nCapacity == 64) {
      // Only 64kbit/s connections are supported by this demo program
      for (i = 0; i < OTXDSP_SDK_WORDSLOTS_PER_HIGHWAY; i++) {
        if (pInS->sqConnectMask[i]) {
          break;
        }
      }

      if (pInS->nDirection == OTXDSP_SDK_RX_DIRECTION) {
        // Connection data for the Receiver
        g_nRxWordSlot = i;
        // Even or odd timeslot
        g_bRxEvenTimeSlot = (pInS->sqConnectMask[i] == 0x00FF) ? OTXDSP_TRUE : OTXDSP_FALSE;

#if 1 // just while testing
        g_nTxWordSlot = g_nRxWordSlot;
        g_bTxEvenTimeSlot = g_bRxEvenTimeSlot;
#endif
      } else {
        // Connection data for the Transmitter
        g_nTxWordSlot = i;
        // Even or odd timeslot
        g_bTxEvenTimeSlot = (pInS->sqConnectMask[i] == 0x00FF) ? OTXDSP_TRUE : OTXDSP_FALSE;
      }

      //g_debugDataS.nData1 = g_nRxWordSlot;
      //g_debugDataS.nData2 = g_bRxEvenTimeSlot;
      
#if 1 // Temporary since we dont get ENABLE yet
      g_bRxEnabled = OTXDSP_TRUE;
      g_bTxEnabled = OTXDSP_TRUE;
#endif

      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);
    } else {
      // Wrong Pipe capacity. Only 64kbit/s connections are supported by this demo program
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_INVALIDARG);
    }


  } else if (nCode == OTXDSP_SDK_IO_DISCONNECT) {
    // No specific action. Just disable.
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_RESET) {
    // No special reset action in this example. Just clear the debug data struct
    g_debugDataS.nData0 = 0;
    g_debugDataS.nData1 = 0;
    g_debugDataS.nData2 = 0;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_INIT_BSP) {
    OtxDspInitBufferedSerialPorts(); 
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_START_PROGRAM) {
#if 0 // Old way is still supported by the driver
    unsigned short dspRc;
    dspRc = OtxDspStartProgram((OtxDspSdkIoInStartProgramS *)OTXDSP_SDK_IOCTL_DATA_ADDR);
    OtxDspCommandCompleteRc(hHandle, dspRc);
#else // Provide more information back to the host driver upon startup
    OtxDspSdkIoOutStartProgramS sOut;

    sOut.nResult = OtxDspStartProgram((OtxDspSdkIoInStartProgramS *)OTXDSP_SDK_IOCTL_DATA_ADDR);

    sOut.nEventBufAddr     = OTXDSP_SDK_EVENT_BUF_ADDR;
    sOut.nEventBufLen      = OTXDSP_SDK_EVENT_BUF_LEN;
    sOut.nIoctlCodeAddr    = OTXDSP_SDK_IOCTL_CODE_ADDR;
    sOut.nIoctlDataMaxSize = 0; //(unsigned short)(OTXDSP_SDK_IOCTL_DATA_MAX_SIZE);

    sOut.nReserved0 = sOut.nReserved1 = sOut.nReserved2 = sOut.nReserved3 = sOut.nReserved4 = 0;

    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(OtxDspSdkIoOutStartProgramS));
#endif
    
  } else if (nCode == OTXDSP_SDK_IO_INIT_TIMER) {
    unsigned short dspRc;
    dspRc = OtxDspInitTimer();
    OtxDspCommandCompleteRc(hHandle, dspRc);
    
  } else if (nCode == OTXDSP_SDK_IO_GET_FRAME_BUFFER_COUNTER) {
    OtxDspSdkIoOutGetFrameBufferCounter sOut;
    sOut.nResult = OTXDSP_SDK_S_OK;
    sOut.nFrameBufferCounter = g_nFrameBufferCounter;
    sOut.nReserved0          = g_nFrameCounter;
    sOut.nReserved1          = 0;
    sOut.nReserved2          = 0;
    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(OtxDspSdkIoOutGetFrameBufferCounter));

    

  // Application specific IO Control Codes
  } else if (nCode == DEMO2_IOCTL_SET_TX_GAIN) {
    Demo2IoInSetTxGainS *pInS = (Demo2IoInSetTxGainS *)OTXDSP_SDK_IOCTL_DATA_ADDR;

    g_nTxGain = pInS->nTxGain;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_STREAM_RX_DATA) {
    Demo2IoInStreamRxDataS *pInS = (Demo2IoInStreamRxDataS *)OTXDSP_SDK_IOCTL_DATA_ADDR;

    g_bStreamRxData = pInS->bStreamRxData;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_STREAM_FULL_FRAMES) {
    Demo2IoInStreamFullFramesS *pInS = (Demo2IoInStreamFullFramesS *)OTXDSP_SDK_IOCTL_DATA_ADDR;

    g_bStreamFullFrames = pInS->bStreamFullFrames;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_GET_SAMPLE) {
    Demo2IoOutGetSampleS sOut;

    sOut.nResult = OTXDSP_SDK_S_OK;

    sOut.nDataSample = g_sqRxData[0];

    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(Demo2IoOutGetSampleS));

  } else if (nCode == DEMO2_IOCTL_GET_FULL_FRAME) {
    Demo2IoOutGetFullFrameS sOut;
    short i;

    sOut.nResult = OTXDSP_SDK_S_OK;

    for (i = 0; i < g_nWordslotsPerFrame*2; i+=2) {
      sOut.nFrame[i+0] = g_sqFrameSample[i/2] >> 8;
      sOut.nFrame[i+1] = g_sqFrameSample[i/2] & 0x00FF;
    }

    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(Demo2IoOutGetFullFrameS));

  } else if (nCode == DEMO2_IOCTL_GET_DEBUG_DATA) {

    // This is an example of a non-blocking call
    // We must confirm the IoControl command before we can send any other events
    // back to the host.
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

    // A non-blocking call would normally take some time to execute.
    // Normally the application should save the Task Reference value, perform the task,
    // then call OtxDspSetHostEvent using the save Task Reference value.
    // However, in this example we call OtxDspSetHostEvent immediately.

    g_debugDataS.nData0 = g_debugData[0];
    g_debugDataS.nData1 = g_debugData[1];
    g_debugDataS.nData2 = g_debugData[2];
    g_debugDataS.nData3 = g_debugData[3];
    g_debugDataS.nData4 = g_debugData[4];
    g_debugDataS.nData5 = g_debugData[5];
    g_debugDataS.nData6 = g_debugData[6];
    g_debugDataS.nData7 = g_debugData[7];

#if 0
    {
      short i;
      for (i = 0; i < 32; i++) {
        g_debugDataS.sqBuf[i] = g_debugBuf[i];
      }
    }
#endif
#if 0
    {
      short i;
      unsigned short *pRxData;
      static short g_index = 0;
      pRxData = (unsigned short *) (OTXDSP_BSP_TXBUF_ADDR + g_index*32);
      for (i = 0; i < 32; i++) {
        g_debugDataS.sqBuf[i] = *pRxData++;
      }
      g_index++;
    }
#endif
#if 0
  // Test cases for G.711. Only for debugging/testing
  {
    memset(&g_debugDataS.sqBuf[0], 0, 32);
    
    g_debugDataS.sqBuf[0] = OtxDspUlawCompress(-2460); // 0x1C
    g_debugDataS.sqBuf[1] = OtxDspUlawCompress(-1505); // 0x27
    g_debugDataS.sqBuf[2] = OtxDspUlawCompress(-650);  // 0x3A
    g_debugDataS.sqBuf[3] = OtxDspUlawCompress(-338);  // 0x48
    g_debugDataS.sqBuf[4] = OtxDspUlawCompress(-90);   // 0x61
    g_debugDataS.sqBuf[5] = OtxDspUlawCompress(-1);    // 0x7E
    g_debugDataS.sqBuf[6] = OtxDspUlawCompress(+102);  // 0xDF
    g_debugDataS.sqBuf[7] = OtxDspUlawCompress(+169);  // 0xD6
    g_debugDataS.sqBuf[8] = OtxDspUlawCompress(+420);  // 0xC3
    g_debugDataS.sqBuf[9] = OtxDspUlawCompress(+499);  // 0xBF
    g_debugDataS.sqBuf[10] = OtxDspUlawCompress(+980); // 0xB0
    g_debugDataS.sqBuf[11] = OtxDspUlawCompress(+7000);// 0x84
    g_debugDataS.sqBuf[12] = 0x1234;

    g_debugDataS.sqBuf[16] = OtxDspLawExpand(0x1C, OTXDSP_ULAW);  //  approx -2460 (0xF664)
    g_debugDataS.sqBuf[17] = OtxDspLawExpand(0x27, OTXDSP_ULAW);  //  approx -1505 (0xFA1F)
    g_debugDataS.sqBuf[18] = OtxDspLawExpand(0x3A, OTXDSP_ULAW);  //  approx -650  (0xFD76)
    g_debugDataS.sqBuf[19] = OtxDspLawExpand(0x48, OTXDSP_ULAW);  //  approx -338  (0xFEAE)
    g_debugDataS.sqBuf[20] = OtxDspLawExpand(0x61, OTXDSP_ULAW);  //  approx -90   (0xFFA6)
    g_debugDataS.sqBuf[21] = OtxDspLawExpand(0x7E, OTXDSP_ULAW);  //  approx -1    (0xFFFF)
    g_debugDataS.sqBuf[22] = OtxDspLawExpand(0xDF, OTXDSP_ULAW);  //  approx +102  (0x0066)
    g_debugDataS.sqBuf[23] = OtxDspLawExpand(0xD6, OTXDSP_ULAW);  //  approx +169  (0x00A9)
    g_debugDataS.sqBuf[24] = OtxDspLawExpand(0xC3, OTXDSP_ULAW);  //  approx +420  (0x01A4)
    g_debugDataS.sqBuf[25] = OtxDspLawExpand(0xBF, OTXDSP_ULAW);  //  approx +499  (0x01F3)
    g_debugDataS.sqBuf[26] = OtxDspLawExpand(0xB0, OTXDSP_ULAW);  //  approx +980  (0x03D4)
    g_debugDataS.sqBuf[27] = OtxDspLawExpand(0x84, OTXDSP_ULAW);  //  approx +7000 (0x1B58)
  }
#endif
#if 0
  // Test cases for G.711. Only for debugging/testing
  {
    memset(&g_debugDataS.sqBuf[0], 0, 32);

    g_debugDataS.sqBuf[0] = OtxDspAlawCompress(-2460*2); // 0xA6 or 0x26
    g_debugDataS.sqBuf[1] = OtxDspAlawCompress(-1505*2); // 0xB2 or 0x32
    g_debugDataS.sqBuf[2] = OtxDspAlawCompress(-650*2);  // 0x81 or 0x01
    g_debugDataS.sqBuf[3] = OtxDspAlawCompress(-338*2);  // 0x90 or 0x10
    g_debugDataS.sqBuf[4] = OtxDspAlawCompress(-90*2);   // 0xF3 or 0x73
    g_debugDataS.sqBuf[5] = OtxDspAlawCompress(-1*2);    // 0xD5 or 0x55
    g_debugDataS.sqBuf[6] = OtxDspAlawCompress(+40*2);   // 0x41 or 0xC1
    g_debugDataS.sqBuf[7] = OtxDspAlawCompress(+102*2);  // 0x7C or 0xFC
    g_debugDataS.sqBuf[8] = OtxDspAlawCompress(+169*2);  // 0x60 or 0xE0
    g_debugDataS.sqBuf[9] = OtxDspAlawCompress(+420*2);  // 0x1F or 0x9F
    g_debugDataS.sqBuf[10] = OtxDspAlawCompress(+499*2); // 0x1A or 0x9A
    g_debugDataS.sqBuf[11] = OtxDspAlawCompress(+980*2); // 0x0B or 0x8B
    g_debugDataS.sqBuf[12] = 0xCCDD;
    
    g_debugDataS.sqBuf[16] = OtxDspLawExpand(0xA6 ^ 0x80, OTXDSP_ALAW);  //  approx -4920 (0xECC8)
    g_debugDataS.sqBuf[17] = OtxDspLawExpand(0xB2 ^ 0x80, OTXDSP_ALAW);  //  approx -3010 (0xF43E)
    g_debugDataS.sqBuf[18] = OtxDspLawExpand(0x81 ^ 0x80, OTXDSP_ALAW);  //  approx -1300 (0xFAEC)
    g_debugDataS.sqBuf[19] = OtxDspLawExpand(0x90 ^ 0x80, OTXDSP_ALAW);  //  approx -676  (0xFD5C)
    g_debugDataS.sqBuf[20] = OtxDspLawExpand(0xF3 ^ 0x80, OTXDSP_ALAW);  //  approx -180  (0xFF4C)
    g_debugDataS.sqBuf[21] = OtxDspLawExpand(0xD5 ^ 0x80, OTXDSP_ALAW);  //  approx -2    (0xFFFE)
    g_debugDataS.sqBuf[22] = OtxDspLawExpand(0x41 ^ 0x80, OTXDSP_ALAW);  //  approx +80   (0x0050)
    g_debugDataS.sqBuf[23] = OtxDspLawExpand(0x7C ^ 0x80, OTXDSP_ALAW);  //  approx +204  (0x00CC)
    g_debugDataS.sqBuf[24] = OtxDspLawExpand(0x60 ^ 0x80, OTXDSP_ALAW);  //  approx +338  (0x0152)
    g_debugDataS.sqBuf[25] = OtxDspLawExpand(0x1F ^ 0x80, OTXDSP_ALAW);  //  approx +840  (0x0348)
    g_debugDataS.sqBuf[26] = OtxDspLawExpand(0x1A ^ 0x80, OTXDSP_ALAW);  //  approx +998  (0x03E6)
    g_debugDataS.sqBuf[27] = OtxDspLawExpand(0x0B ^ 0x80, OTXDSP_ALAW);  //  approx +1960 (0x07A8)
    
  }
#endif
#if 0
  // Test cases for external memory. Only for debugging/testing
  {
    short i;
    memset(&g_debugDataS.sqBuf[0], 0, 32);

    for (i= 0; i < 256; i++) {
      g_sqTestData[i] = i;
    }
    g_debugDataS.sqBuf[0] = g_sqTestData[0];
    g_debugDataS.sqBuf[1] = g_sqTestData[1];
    g_debugDataS.sqBuf[2] = g_sqTestData[2];
    g_debugDataS.sqBuf[31] = g_sqTestData[255];
    
  }
#endif
#if 0
  // Test cases for external memory. Only for debugging/testing
  {
    short i;
    memset(&g_debugDataS.sqBuf[0], 0, 32);

    g_debugDataS.sqBuf[0] = Foo();
    g_debugDataS.sqBuf[31] = 10;
    
  }
#endif


    OtxDspSetHostEvent(DEMO2_NOTIFY_DEBUG_DATA, hHandle, nTaskRef, (unsigned short *)&g_debugDataS, sizeof(Demo2NotifyGetDebugDataS));

  } else if (nCode == DEMO2_IOCTL_SET_BLINK_FREQ) {
    Demo2IoInSetBlinkFreqS *pInS = (Demo2IoInSetBlinkFreqS *)OTXDSP_SDK_IOCTL_DATA_ADDR;

    g_blinkCountMax = pInS->nBlinkCountFactor * 10000L;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_CONST_VAL_MODE_ENABLE) {
    short n;
    unsigned short *pTxData;

    g_bConstValMode = OTXDSP_TRUE;

    // set to transmit idle character
    pTxData = (unsigned short *) (OTXDSP_BSP_TXBUF_ADDR);
    for (n = 0; n < OTXDSP_BSP_TXBUF_SIZE; n++) {
      *pTxData++ = 0xffff;
    }

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_CONST_VAL_MODE_DISABLE) {
    g_bConstValMode = OTXDSP_FALSE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_SET_CONST_VAL) {
    Demo2IoInSetConstValS *pInS = (Demo2IoInSetConstValS *) OTXDSP_SDK_IOCTL_DATA_ADDR;

    if (pInS->nTimeSlot < 32) {
      short n;
      short c;

      if (pInS->bSet) {
        c = pInS->nConstVal & 0xff;
      } else {
        c = 0xff;   // idle character
      }

      // set to transmit constant value
      for (n = 0; n < OTXDSP_BSP_TXBUF_SIZE; n += 16) {
        unsigned short *pTxData;

        pTxData = (unsigned short *) (OTXDSP_BSP_TXBUF_ADDR + n + (pInS->nTimeSlot / 2));
        if (pInS->nTimeSlot % 2) {    // odd
          *pTxData = (*pTxData & 0xff00) | (c); 
        } else {    // even
          *pTxData = (*pTxData & 0x00ff) | (c << 8);
        }
      }

      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);
    } else {
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_INVALIDARG);
    }

  } else {
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_INVALIDARG);
  }

}


void OtxDspServiceRxSerialPort(
/*--------------------------------------------------------------------------*\
    Call-back routine call when we have receive interrupt from the
    Buffered Serial Port indicating the the receive buffer is either full or
    half-full.

    This function is called from the idle loop in main()
    TO BE RE-IMPLEMENTED BY THE USER.
\*--------------------------------------------------------------------------*/
  void
)
{
  short n;
  unsigned short *pTxData;
  unsigned short *pRxData;

  g_nFrameCounter += DEMO_DATA_BUF_SIZE;

  if (g_bRxSerialPortHalf) {
    // BSPCE RH=1 --> We read the second half
    pRxData = (unsigned short *) (OTXDSP_BSP_RXBUF_ADDR + OTXDSP_BSP_RXBUF_SIZE/2);
    pTxData = (unsigned short *) (OTXDSP_BSP_TXBUF_ADDR + OTXDSP_BSP_TXBUF_SIZE/2);
  } else {
    // BSPCE RH=0 --> We read the first half
    pRxData = (unsigned short *) (OTXDSP_BSP_RXBUF_ADDR);
    pTxData = (unsigned short *) (OTXDSP_BSP_TXBUF_ADDR);
  }

  // Copy one full frame into a temporary storage buffer
  // Used by the DEMO2_IOCTL_GET_FULL_FRAME command
  memcpy(&g_sqFrameSample[0], pRxData, g_nWordslotsPerFrame);

  if (g_bStreamFullFrames) {
    // Send all received data as a stream to the host
    static unsigned short nTaskRef = 0;
    OtxDspSetHostEvent(DEMO2_NOTIFY_STREAM_FULL_FRAMES, g_hHandle, nTaskRef++, pRxData, OTXDSP_BSP_RXBUF_SIZE/2);
  }

  // At this point there are DEMO_DATA_BUF_SIZE frames of data to pick up
  // It is also time to fill the transmit buffer with data for DEMO_DATA_BUF_SIZE frames
  if (!g_bConstValMode) {
    g_iTxData = 0;
    g_iRxData = 0;
    for (n = 0; n < DEMO_DATA_BUF_SIZE; n++) {
      if (g_bRxEnabled) {
        UnpackFrame(pRxData + n*g_nWordslotsPerFrame);  // There are 16 words in one frame
      }
      if (g_bTxEnabled) {
        PackFrame(pTxData + n*g_nWordslotsPerFrame);
      }
    }
  }

  ProcessRxSamples();
  GenerateTxSamples();

}


static void UnpackFrame(
/*--------------------------------------------------------------------------*\
  @func
    Picks up a certain timeslot from a frame, then expands the 8-bit timeslot
    to a 16-bit value using either u-law or a-law expansion (14-bit value with
    the 2 LSB set to 0).

    This function is called from OtxDspServiceRxSerialPort().

    Note that the TS0 octet is the most significant octet of the first
    unsigned short (16-bit) value in the frame. The TS1 octet is the least
    significant octet of the first unsigned short (16-bit) value in the frame.
    Therefore we must an 8 bit shift depending on if we want to pick up an even
    or an odd timeslot octet.

\*--------------------------------------------------------------------------*/
  unsigned short *pFrame
)
{
  //short n;
  unsigned short nTimeSlotOctet;

  //g_debugData[0]++;


  nTimeSlotOctet = *(pFrame + g_nRxWordSlot);
  if (g_bRxEvenTimeSlot) {
    // If we are working with an even timeslot (0,2,...,28,30) we must shift it down
    // to the Least Significant octet
    nTimeSlotOctet >>= 8;
  } else {
    // Odd timeslot. Just ignore the upper 8 bits to get the timeslot octet
    nTimeSlotOctet &= 0x00FF;
  }

  // Expand according to current law and store the data in our receive buffer
#if 0
  g_sqRxData[g_iRxData++] = OtxDspLawExpand(nTimeSlotOctet, g_law);
#else // Just for testing
  g_sqRxData[g_iRxData++] = nTimeSlotOctet;
#endif
}


static void PackFrame(
/*--------------------------------------------------------------------------*\
  @func

    Performs law compression on the outgoing data values. The compressed octet
    is then stored in the corresponding timeslot in the frame.

    This function is called from OtxDspServiceRxSerialPort().
    
\*--------------------------------------------------------------------------*/
  unsigned short *pFrame
)
{
  //short n;
  //g_debugData[1]++;


  if (g_law == OTXDSP_ALAW) {
    // A-law
    if (g_bTxEvenTimeSlot) {
      // Even time slot. OR in the 8 bit compressed value in 8 MSB bits of the frame Word
      *(pFrame + g_nTxWordSlot) &= 0x00FF;
      *(pFrame + g_nTxWordSlot) |= ((OtxDspAlawCompress(g_sqTxData[g_iTxData++])) << 8);
    } else {
      // Odd time slot. OR in the 8 bit compressed value in 8 LSB bits of the frame Word
      *(pFrame + g_nTxWordSlot) &= 0xFF00;
      *(pFrame + g_nTxWordSlot) |= OtxDspAlawCompress(g_sqTxData[g_iTxData++]);
    }
  } else {
    // U-law
    if (g_bTxEvenTimeSlot) {
      // Even time slot. OR in the 8 bit compressed value in 8 MSB bits of the bch Word
      *(pFrame + g_nTxWordSlot) &= 0x00FF;
#if 1
      *(pFrame + g_nTxWordSlot) |= ((OtxDspUlawCompress(g_sqTxData[g_iTxData++])) << 8);
#else // Just for testing
      *(pFrame + g_nTxWordSlot) |= (0x0b << 8);
#endif
    } else {
      // Odd time slot. OR in the 8 bit compressed value in 8 LSB bits of the bch Word
      *(pFrame + g_nTxWordSlot) &= 0xFF00;
#if 1
      *(pFrame + g_nTxWordSlot) |= OtxDspUlawCompress(g_sqTxData[g_iTxData++]);
#else // Just for testing
      *(pFrame + g_nTxWordSlot) |= 0xcd;
#endif
    }
  }

}



void OtxDspServiceTxSerialPort(
/*--------------------------------------------------------------------------*\
    Call-back routine call when we have transmit interrupt from the
    Buffered Serial Port indicating the the transmit buffer is either empty or
    half-empty.

    This function is called from the idle loop in main()
    TO BE RE-IMPLEMENTED BY THE USER.
\*--------------------------------------------------------------------------*/
  void
)
{
  // Performed by OtxDspServiceRxSerialPort() in this example
}


void ProcessRxSamples(
/*--------------------------------------------------------------------------*\
    Process Received Samples
\*--------------------------------------------------------------------------*/
  void
)
{
  if (g_bStreamRxData) {
    // Send the received data stream to the host
    static unsigned short nTaskRef = 0;
    Demo2NotifyStreamRxDataS streamRxDataS;

    memcpy(streamRxDataS.nData, g_sqRxData, 25);
    OtxDspSetHostEvent(DEMO2_NOTIFY_STREAM_RX_DATA, g_hHandle, nTaskRef++, (unsigned short *)&streamRxDataS, sizeof(Demo2NotifyStreamRxDataS));
  }
}


void GenerateTxSamples(
/*--------------------------------------------------------------------------*\
    Generate more samples to Transmit
\*--------------------------------------------------------------------------*/
  void
)
{
  short i;

  // Simply copy the received samples to the outgoing transmit data buffer
  // Perform a gain adjustment for each copied sample

  for (i = 0; i < DEMO_DATA_BUF_SIZE; i++) {
    g_sqTxData[i] = g_sqRxData[i] << g_nTxGain;
  }

}

